﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Visualization;

using Rectangle = HeuristicLab.Visualization.Rectangle;

namespace HeuristicLab.Encodings.SymbolicExpressionTreeEncoding.Views {
  /// <summary>
  /// The SymbolicExpressionTreeTile will display either a full tree or just a subtree (case in which the SymbolicExpressionTreeProperty remains unset).
  /// By setting the Root property, the user can use the SymbolicExpressionTreeTile to layout and display arbitrary subtrees.
  /// </summary>
  public class SymbolicExpressionTreeTile : Group {
    public double PreferredNodeWidth { get; set; }
    public double PreferredNodeHeight { get; set; }

    public Size Size {
      get {
        int xmin = 0, ymin = 0, xmax = 0, ymax = 0;

        foreach (var p in Primitives.OfType<RectangularPrimitiveBase>()) {
          if (xmin > p.LowerLeft.X) xmin = (int)Math.Floor(p.LowerLeft.X);
          if (xmax < p.UpperRight.X) xmax = (int)Math.Ceiling(p.UpperRight.X);
          if (ymin > p.LowerLeft.Y) ymin = (int)Math.Floor(p.LowerLeft.Y);
          if (ymax < p.UpperRight.Y) ymax = (int)Math.Ceiling(p.UpperRight.Y);
        }

        return new Size(xmax - xmin, ymax - ymin);
      }
    }

    private Point position;
    public Point Position {
      get { return position; }
      set {
        var oldpos = position;
        position = value;
        int ox = position.X - oldpos.X;
        int oy = position.Y - oldpos.Y;
        // translate all primitives to the new position
        foreach (var p in Primitives) {
          var rpb = p as RectangularPrimitiveBase;
          if (rpb != null) {
            var lowerLeft = new Point((int)Math.Floor(rpb.LowerLeft.X) + ox, (int)Math.Floor(rpb.LowerLeft.Y + oy));
            var upperRight = new PointD(lowerLeft.X + rpb.Size.Width, lowerLeft.Y + rpb.Size.Height);
            rpb.SetPosition(lowerLeft, upperRight);
          }
          var line = p as LinearPrimitiveBase;
          if (line != null) {
            var start = new Point((int)Math.Floor(line.Start.X) + ox, (int)Math.Floor(line.Start.Y) + oy);
            var end = new Point(start.X + (int)line.Size.Width, start.Y + (int)line.Size.Height);
            line.SetPosition(start, end);
          }
        }
      }
    }

    private ISymbolicExpressionTree symbolicExpressionTree;
    public ISymbolicExpressionTree SymbolicExpressionTree {
      get { return symbolicExpressionTree; }
      set {
        symbolicExpressionTree = value;
        Root = symbolicExpressionTree.Root;
      }
    }

    new private void Clear() {
      Group.Clear();
      primitiveMap.Clear();
    }

    private ISymbolicExpressionTreeNode root;
    public ISymbolicExpressionTreeNode Root {
      get { return root; }
      set {
        root = value;
        GeneratePrimitives(PreferredNodeWidth, PreferredNodeHeight);
      }
    }
    public ReingoldTilfordLayoutEngine<ISymbolicExpressionTreeNode> LayoutEngine { get; set; }
    private readonly Dictionary<IPrimitive, ISymbolicExpressionTreeNode> primitiveMap;

    public SymbolicExpressionTreeTile(IChart chart)
      : base(chart) {
      primitiveMap = new Dictionary<IPrimitive, ISymbolicExpressionTreeNode>();
      PreferredNodeWidth = 80;
      PreferredNodeHeight = 40;
      Group = new Group(chart);
    }
    public SymbolicExpressionTreeTile(IChart chart, ISymbolicExpressionTree tree)
      : this(chart) {
      SymbolicExpressionTree = tree;
    }
    private void GeneratePrimitives(double preferredNodeWidth, double preferredNodeHeight) {
      Clear();
      var actualRoot = Root;
      if (Root.Symbol is ProgramRootSymbol && Root.SubtreeCount == 1) { actualRoot = Root.GetSubtree(0); }
      var visualNodes = LayoutEngine.CalculateLayout(actualRoot);

      var font = new Font(FontFamily.GenericSansSerif, 10, GraphicsUnit.Pixel);

      var visualNodeMap = visualNodes.ToDictionary(x => x.Content, x => x);

      foreach (var visualNode in visualNodes) {
        var lowerLeft = new PointD(visualNode.X, visualNode.Y);
        var upperRight = new PointD(visualNode.X + preferredNodeWidth, visualNode.Y + preferredNodeHeight);
        var node = visualNode.Content;
        RectangularPrimitiveBase rectangularPrimitive;
        if (node.SubtreeCount == 0) {
          rectangularPrimitive = new Rectangle(Chart, lowerLeft, upperRight) { Font = font, Text = visualNode.Content.ToString() };
        } else {
          rectangularPrimitive = new Ellipse(Chart, lowerLeft, upperRight) { Font = font, Text = visualNode.Content.ToString() };
        }

        primitiveMap.Add(rectangularPrimitive, node); // to be able to retrieve nodes via primitives
        this.Add(rectangularPrimitive);

        //        int x = Position.X, y = Position.Y;
        //        if (x > rectangularPrimitive.LowerLeft.X) {
        //          x = (int)Math.Floor(rectangularPrimitive.LowerLeft.X);
        //        }
        //        if (y > rectangularPrimitive.LowerLeft.Y) {
        //          y = (int)Math.Floor(rectangularPrimitive.LowerLeft.Y);
        //        }
        //        Position = new Point(x, y);
        //        int w = Size.Width, h = Size.Height;


        if (rectangularPrimitive.Size.Width.IsAlmost(0) || rectangularPrimitive.Size.Height.IsAlmost(0)) {
          throw new Exception("Primitive size cannot be zero.");
        }
      }

      foreach (var node in visualNodes.Where(n => n.Content.SubtreeCount > 0)) {
        var parent = node.Content;
        foreach (var child in parent.Subtrees.Select(x => visualNodeMap[x])) {
          var start = new PointD(node.X + preferredNodeWidth / 2, node.Y + preferredNodeHeight);
          var end = new PointD(child.X + preferredNodeWidth / 2, child.Y);
          var line = new Line(this.Chart, start, end);
          this.Add(line);
        }
      }
    }
  }
}
